@@ -6,14 +6,23 @@
 #include <memory>
 #include <sstream>
 #include <stdexcept>
+#include <type_traits>
 #include <unordered_set>
 #include <vector>
 
+#include "big_int.h"
+#include "utils.h"
+
 namespace py = pybind11;
 
+#ifndef Py_SET_SIZE
+#define Py_SET_SIZE(object, size) ((PyVarObject*)(object))->ob_size = size
+#endif
+
 #define MODULE_NAME cppbuiltins
 #define C_STR_HELPER(a) #a
 #define C_STR(a) C_STR_HELPER(a)
+#define INT_NAME "int"
 #define LIST_ITERATOR_NAME "list_iterator"
 #define LIST_NAME "list"
 #define LIST_REVERSED_ITERATOR_NAME "list_reversed_iterator"
@@ -93,13 +102,146 @@ static bool operator==(const Object& left, const Object& right) {
 }  // namespace pybind11
 
 template <class Type>
-std::string to_repr(const Type& value) {
+static std::string to_repr(const Type& value) {
   std::ostringstream stream;
   stream.precision(std::numeric_limits<double>::digits10 + 2);
   stream << value;
   return {stream.str()};
 }
 
+using BaseInt =
+    BigInt<std::conditional_t<sizeof(void*) == 8, std::uint32_t, std::uint16_t>,
+           '_'>;
+
+static int int_to_sign(const py::int_& value) {
+  PyLongObject* ptr = (PyLongObject*)value.ptr();
+  Py_ssize_t signed_size = Py_SIZE(ptr);
+  return signed_size < 0 ? -1 : signed_size > 0;
+}
+
+static std::vector<BaseInt::Digit> int_to_digits(const py::int_& value) {
+  PyLongObject* ptr = (PyLongObject*)value.ptr();
+  Py_ssize_t signed_size = Py_SIZE(ptr);
+  std::size_t size = Py_ABS(signed_size) + (signed_size == 0);
+  std::vector<BaseInt::Digit> result =
+      binary_digits_to_binary_base<digit, BaseInt::Digit, PyLong_SHIFT,
+                                   BaseInt::BINARY_SHIFT>(
+          std::vector<digit>(ptr->ob_digit, ptr->ob_digit + size));
+  return result;
+}
+
+static const char* pystr_to_ascii_c_str(const py::str& string) {
+  py::str ascii_string = py::reinterpret_steal<py::str>(
+      _PyUnicode_TransformDecimalAndSpaceToASCII(string.ptr()));
+  if (!ascii_string) throw py::error_already_set();
+  const char* result = PyUnicode_AsUTF8(ascii_string.ptr());
+  if (!result) throw py::error_already_set();
+  return result;
+}
+
+class Int : public BaseInt {
+ public:
+  Int() : BaseInt() {}
+
+  explicit Int(const BaseInt& value) : BaseInt(value) {}
+
+  explicit Int(py::int_ value)
+      : BaseInt(int_to_sign(value), int_to_digits(value)) {}
+
+  Int(const py::str& value, std::size_t base)
+      : BaseInt(pystr_to_ascii_c_str(value), base) {}
+
+  static Int from_state(const py::int_& value) { return Int(value); }
+
+  static py::int_ to_state(const Int& value) {
+    return py::reinterpret_steal<py::int_>((PyObject*)value.as_PyLong());
+  }
+
+  Int operator%(const Int& divisor) const {
+    Int quotient, result;
+    divmod(divisor, quotient, result);
+    return result;
+  }
+
+  const Int& operator+() const { return *this; }
+
+  Int operator+(const Int& other) const {
+    return Int(BaseInt::operator+(other));
+  }
+
+  Int operator~() const { return Int(BaseInt::operator~()); }
+
+  Int operator*(const Int& other) const {
+    return Int(BaseInt::operator*(other));
+  }
+
+  Int operator-() const { return Int(BaseInt::operator-()); }
+
+  Int operator-(const Int& other) const {
+    return Int(BaseInt::operator-(other));
+  }
+
+  Int abs() const { return Int(BaseInt::abs()); }
+
+  PyLongObject* as_PyLong() const {
+    int sign = this->sign();
+    const std::vector<BaseInt::Digit>& digits = this->digits();
+    std::vector<digit> result_digits = binary_digits_to_binary_base<
+        BaseInt::Digit, digit, BaseInt::BINARY_SHIFT, PyLong_SHIFT>(digits);
+    PyLongObject* result = _PyLong_New(result_digits.size());
+    std::memcpy(result->ob_digit, result_digits.data(),
+                sizeof(digit) * result_digits.size());
+    Py_SET_SIZE(result, Py_SIZE(result) * sign);
+    return result;
+  }
+
+  void divmod(const BigInt& divisor, BigInt& quotient,
+              BigInt& remainder) const {
+    try {
+      BaseInt::divmod(divisor, quotient, remainder);
+    } catch (const std::range_error& exception) {
+      PyErr_SetString(PyExc_ZeroDivisionError, exception.what());
+      throw py::error_already_set();
+    }
+  }
+
+  Int floor_divide(const Int& divisor) const {
+    Int result, remainder;
+    divmod(divisor, result, remainder);
+    return result;
+  }
+
+  Int gcd(const Int& other) const { return Int(BaseInt::gcd(other)); }
+
+  Py_hash_t hash() const {
+    int sign = this->sign();
+    const std::vector<BaseInt::Digit>& digits = this->digits();
+    if (digits.size() == 1) {
+      if (sign > 0)
+        return digits[0];
+      else if (sign < 0)
+        return -static_cast<Py_hash_t>(digits[0] + (digits[0] == 1));
+      else
+        return 0;
+    }
+    Py_uhash_t result = 0;
+    for (auto position = digits.rbegin(); position != digits.rend();
+         ++position) {
+      result = ((result << BaseInt::BINARY_SHIFT) & _PyHASH_MODULUS) |
+               (result >> (_PyHASH_BITS - BaseInt::BINARY_SHIFT));
+      result += *position;
+      if (result >= _PyHASH_MODULUS) result -= _PyHASH_MODULUS;
+    }
+    result = result * sign;
+    result -= (result == std::numeric_limits<std::size_t>::max());
+    return static_cast<Py_hash_t>(result);
+  }
+};
+
+static std::ostream& operator<<(std::ostream& stream, const Int& value) {
+  return stream << C_STR(MODULE_NAME) "." INT_NAME "('" << value.repr() << "')";
+}
+
 template <class Iterable>
 IterableState iterable_to_state(const Iterable& self) {
   IterableState result;
@@ -238,9 +380,7 @@ class List {
                                         other._raw->begin(), other._raw->end());
   }
 
-  bool operator<=(const List& other) const {
-    return *this < other || *this == other;
-  }
+  bool operator<=(const List& other) const { return !(other < *this); }
 
   List operator+(const List& other) const {
     RawList raw{*_raw};
@@ -458,7 +598,7 @@ class List {
       const Size size = values.size();
       RawList keys;
       keys.reserve(size);
-      for (const auto value : values) keys.push_back(key(value));
+      for (const auto& value : values) keys.push_back(key(value));
       std::vector<Size> indices;
       indices.reserve(size);
       for (Size index = 0; index < size; ++index) indices.push_back(index);
@@ -855,7 +995,45 @@ static std::ostream& operator<<(std::ostream& stream, const Set& set) {
 PYBIND11_MODULE(MODULE_NAME, m) {
   m.doc() =
       R"pbdoc(Alternative implementation of python builtins based on C++ `std` library.)pbdoc";
-  m.attr("__version__") = VERSION_INFO;
+  m.attr("__version__") = C_STR(VERSION_INFO);
+
+  m.def("gcd", &Int::gcd);
+
+  py::class_<Int> PyInt(m, INT_NAME);
+  PyInt.def(py::init<>())
+      .def(py::init<const py::str&, std::size_t>(), py::arg("string"),
+           py::arg("base") = 10)
+      .def(py::init<const py::int_&>(), py::arg("value"))
+      .def(py::self == py::self)
+      .def(py::self <= py::self)
+      .def(py::self < py::self)
+      .def(py::self + py::self)
+      .def(py::self % py::self)
+      .def(~py::self)
+      .def(+py::self)
+      .def(py::self * py::self)
+      .def(-py::self)
+      .def(py::self - py::self)
+      .def(py::pickle(&Int::to_state, &Int::from_state))
+      .def("__abs__", &Int::abs)
+      .def("__hash__", &Int::hash)
+      .def("__bool__", &Int::operator bool)
+      .def("__copy__", [](const Int& self) -> const Int& { return self; })
+      .def(
+          "__divmod__",
+          [](const Int& self, const Int& other) {
+            Int quotient, remainder;
+            self.divmod(other, quotient, remainder);
+            return py::make_tuple(quotient, remainder);
+          },
+          py::is_operator{})
+      .def("__deepcopy__",
+           [](const Int& self, const py::dict&) -> Int { return self; })
+      .def("__int__", &Int::to_state)
+      .def("__float__", &Int::operator double)
+      .def("__floordiv__", &Int::floor_divide, py::is_operator{})
+      .def("__repr__", &to_repr<Int>)
+      .def("__str__", [](const Int& self) { return self.repr(); });
 
   py::class_<List> PyList(m, LIST_NAME);
   PyList.def(py::init<py::iterable>(), py::arg("values"))
@@ -949,6 +1127,8 @@ PYBIND11_MODULE(MODULE_NAME, m) {
       .def("union", &Set::union_)
       .def("update", &Set::update);
 
+  collections_abc.attr("MutableSet").attr("register")(PySet);
+
   py::class_<SetIterator>(m, SET_ITERATOR_NAME)
       .def(
           "__reduce__",
